"""This module deals with lubricants dictionary in the ROSS library."""

from ross.units import Q_

lubricants_dict = {
    "ISOVG32": {
        "liquid_viscosity1": Q_(4.05640e-06, "reyn").to_base_units().m,
        "temperature1": Q_(40.00000, "degC").to_base_units().m,
        "liquid_viscosity2": Q_(6.76911e-07, "reyn").to_base_units().m,
        "temperature2": Q_(100.00000, "degC").to_base_units().m,
        "liquid_density": Q_(873.99629, "kg/m**3").to_base_units().m,
        "liquid_specific_heat": Q_(1948.7995685758851, "J/(kg*degK)").to_base_units().m,
        "liquid_thermal_conductivity": Q_(0.13126, "W/(m*degC)").to_base_units().m,
        "vapour_viscosity": Q_(9e-06, "Pa*s").to_base_units().m,
        "vapour_density": Q_(0.029, "kg/m**3").to_base_units().m,
        "vapour_specific_heat": Q_(1000, "J/(kg*degC)").to_base_units().m,
        "vapour_thermal_conductivity": Q_(0.026, "W/(m*degK)").to_base_units().m,
        "saturation_pressure": Q_(1e5, "Pa").to_base_units().m,
        "surface_tension": Q_(0.04, "N/m").to_base_units().m,
        "molecular_weight": Q_(500, "g/mol").to_base_units().m,
    },
    "ISOVG46": {
        "liquid_viscosity1": Q_(5.757040938820288e-06, "reyn").to_base_units().m,
        "temperature1": Q_(40, "degC").to_base_units().m,
        "liquid_viscosity2": Q_(8.810775697672788e-07, "reyn").to_base_units().m,
        "temperature2": Q_(100, "degC").to_base_units().m,
        "liquid_density": Q_(862.9, "kg/m**3").to_base_units().m,
        "liquid_specific_heat": Q_(1950, "J/(kg*degK)").to_base_units().m,
        "liquid_thermal_conductivity": Q_(0.15, "W/(m*degC)").to_base_units().m,
        "vapour_viscosity": Q_(9e-06, "Pa*s").to_base_units().m,
        "vapour_density": Q_(0.029, "kg/m**3").to_base_units().m,
        "vapour_specific_heat": Q_(1000, "J/(kg*degC)").to_base_units().m,
        "vapour_thermal_conductivity": Q_(0.026, "W/(m*degK)").to_base_units().m,
        "saturation_pressure": Q_(1e5, "Pa").to_base_units().m,
        "surface_tension": Q_(0.04, "N/m").to_base_units().m,
        "molecular_weight": Q_(500, "g/mol").to_base_units().m,
    },
    "TEST": {
        "liquid_viscosity1": Q_(0.04, "Pa*s").to_base_units().m,
        "temperature1": Q_(40, "degC").to_base_units().m,
        "liquid_viscosity2": Q_(0.01, "Pa*s").to_base_units().m,
        "temperature2": Q_(100, "degC").to_base_units().m,
        "liquid_density": Q_(863.61302696, "kg/m**3").to_base_units().m,
        "liquid_specific_heat": Q_(1951.88616, "J/(kg*degK)").to_base_units().m,
        "liquid_thermal_conductivity": Q_(0.15, "W/(m*degC)").to_base_units().m,
        "vapour_viscosity": Q_(9e-06, "Pa*s").to_base_units().m,
        "vapour_density": Q_(0.029, "kg/m**3").to_base_units().m,
        "vapour_specific_heat": Q_(1000, "J/(kg*degC)").to_base_units().m,
        "vapour_thermal_conductivity": Q_(0.026, "W/(m*degK)").to_base_units().m,
        "saturation_pressure": Q_(1e5, "Pa").to_base_units().m,
        "surface_tension": Q_(0.04, "N/m").to_base_units().m,
        "molecular_weight": Q_(500, "g/mol").to_base_units().m,
    },
    "ISOVG68": {
        "liquid_viscosity1": Q_(0.0570722, "Pa*s").to_base_units().m,
        "temperature1": Q_(40, "degC").to_base_units().m,
        "liquid_viscosity2": Q_(0.00766498, "Pa*s").to_base_units().m,
        "temperature2": Q_(100, "degC").to_base_units().m,
        "liquid_density": Q_(867, "kg/m**3").to_base_units().m,
        "liquid_specific_heat": Q_(1951, "J/(kg*degK)").to_base_units().m,
        "liquid_thermal_conductivity": Q_(0.15, "W/(m*degC)").to_base_units().m,
        "vapour_viscosity": Q_(9e-06, "Pa*s").to_base_units().m,
        "vapour_density": Q_(0.029, "kg/m**3").to_base_units().m,
        "vapour_specific_heat": Q_(1000, "J/(kg*degC)").to_base_units().m,
        "vapour_thermal_conductivity": Q_(0.026, "W/(m*degK)").to_base_units().m,
        "saturation_pressure": Q_(1e5, "Pa").to_base_units().m,
        "surface_tension": Q_(0.04, "N/m").to_base_units().m,
        "molecular_weight": Q_(500, "g/mol").to_base_units().m,
    },
    "HYDRAXP68": {
        "liquid_viscosity1": Q_(0.0570722, "Pa*s").to_base_units().m,
        "temperature1": Q_(40, "degC").to_base_units().m,
        "liquid_viscosity2": Q_(7.66498e-03, "Pa*s").to_base_units().m,
        "temperature2": Q_(100, "degC").to_base_units().m,
        "liquid_density": Q_(874, "kg/m**3").to_base_units().m,
        "liquid_specific_heat": Q_(2000, "J/(kg*degC)").to_base_units().m,
        "liquid_thermal_conductivity": Q_(0.133, "W/(m*degK)").to_base_units().m,
        "vapour_viscosity": Q_(9e-06, "Pa*s").to_base_units().m,
        "vapour_density": Q_(0.029, "kg/m**3").to_base_units().m,
        "vapour_specific_heat": Q_(1000, "J/(kg*degC)").to_base_units().m,
        "vapour_thermal_conductivity": Q_(0.026, "W/(m*degK)").to_base_units().m,
        "saturation_pressure": Q_(81325, "Pa").to_base_units().m,
        "surface_tension": Q_(0.04, "N/m").to_base_units().m,
        "molecular_weight": Q_(500, "g/mol").to_base_units().m,
    },
    "SEAWATER": {
        "liquid_viscosity1": Q_(1.08e-03, "Pa*s").to_base_units().m,
        "temperature1": Q_(20, "degC").to_base_units().m,
        "liquid_viscosity2": Q_(0.35e-03, "Pa*s").to_base_units().m,
        "temperature2": Q_(100, "degC").to_base_units().m,
        "liquid_density": Q_(1025, "kg/m**3").to_base_units().m,
        "liquid_specific_heat": Q_(3850, "J/(kg*degK)").to_base_units().m,
        "liquid_thermal_conductivity": Q_(0.6, "W/(m*degC)").to_base_units().m,
        "vapour_viscosity": Q_(9.75e-06, "Pa*s").to_base_units().m,
        "vapour_density": Q_(0.597, "kg/m**3").to_base_units().m,
        "vapour_specific_heat": Q_(2010, "J/(kg*degC)").to_base_units().m,
        "vapour_thermal_conductivity": Q_(0.025, "W/(m*degK)").to_base_units().m,
        "saturation_pressure": Q_(101325, "Pa").to_base_units().m,
        "surface_tension": Q_(0.074, "N/m").to_base_units().m,
        "molecular_weight": Q_(18.015, "g/mol").to_base_units().m,
    },
}
